from typing import Callable, List, Type, TypeVar
from django.conf import settings as django_settings
from px_settings import settings as base_settings, update as base_update
from px_settings.types import SettingsType


__all__ = 'collect_prefixed', 'collect_sources', 'update', 'settings',

T = TypeVar('T', bound=Type)
DELIMITER = '_'


# [TODO]: Update px_settings.contrib.django.collect_prefixed to be able to
# resolve from dict, not only an object based resolver.
# Or make a base prefix resolver for dicts and use it in django-related
# version.
def collect_prefixed(obj: object, prefix: str) -> dict:
    strip_by = len(prefix)
    storage = obj._wrapped if hasattr(obj, '_wrapped') else obj

    return {
        key[strip_by:]: value
        for key, value in storage.__dict__.items()
        if key.startswith(prefix)
    }


def collect_sources(
    name: str,
    delimiter: str = DELIMITER,
    settings_module = django_settings
) -> List[dict]:
    assert settings_module.configured is True, 'Configure django settings first.'

    return (
        collect_prefixed(settings_module, name + delimiter),
        getattr(settings_module, name, {}),
    )


def update(
    name: str,
    delimiter: str,
    settings_module,
    settings: T,
    data: dict
) -> T:
    prefixed, from_dict = collect_sources(
        name, delimiter=delimiter, settings_module=settings_module
    )

    return base_update(settings, **{**from_dict, **prefixed, **data})


def settings(
    name: str,
    *sources: List[SettingsType],
    updator: Callable[[T, dict], T] = update,
    delimiter: str = DELIMITER,
    settings_module = django_settings
) -> Callable[[T], T]:
    return base_settings(
        *collect_sources(
            name, delimiter=delimiter, settings_module=settings_module
        ),
        *sources,
        updator=lambda settings, **data: updator(
            name,
            delimiter,
            settings_module,
            settings,
            data
        )
    )



from typing import Dict, Union
import sys


__all__ = 'SettingsType',


SettingsType = Union[dict, tuple]


if sys.version_info >= (3, 8):
    from typing import Protocol


    class DataclassType(Protocol):
        __dataclass_fields__: Dict

    SettingsType = Union[dict, tuple, DataclassType]



import inspect
from typing import Callable, List, Type, TypeVar
from dataclasses import fields, is_dataclass
from itertools import chain
from functools import reduce, partial
import warnings

from .merger import merge_settings_reducer
from .types import SettingsType
from .updator import update


__all__ = 'settings',

empty = object()
T = TypeVar('T', bound=Type)


def settings(
    *base_sources: List[SettingsType],
    updator: Callable[[T, dict], T] = update
) -> Callable[[T], T]:
    """Decorator for settings class wrapping.

    Result will be a callable, that will generate a resulting settings
    object based on a passed sources.
    """

    def wrapper(base: T) -> T:
        flds = {f.name for f in fields(base)} if is_dataclass(base) else None

        def creator(*current_sources: List[SettingsType], **kwargs: SettingsType):
            sources = list(chain([kwargs], current_sources, base_sources))

            assert len(sources) != 0, (
                'You must provide at least one source object to create '
                'settings instance.'
            )
            result = reduce(merge_settings_reducer, sources)

            if flds is not None:
                result = {k: v for k, v in result.items() if k in flds}

            try:
                instance = base(**result)
            except TypeError as e:
                raise TypeError(
                    f'Settings "{base.__module__}.{base.__name__}" has an invalid input: \r\n{e}'
                )

            if getattr(instance, 'update', empty) is not empty:
                current = inspect.stack()[1]

                warnings.warn(ResourceWarning(
                    f'Settings in "{current.filename}" already have `update` value.',
                    'Update mechanics disabled.',
                    'Use raw `update` function to make update if necessary.',
                ))
            else:
                instance.update = partial(updator, instance)

            return instance

        return creator

    return wrapper



from typing import Type, TypeVar


__all__ = 'update',

T = TypeVar('T', bound=Type)


def update(settings: T, **data: dict) -> T:
    for key in data:
        setattr(settings, key, data[key])

    return settings



from dataclasses import is_dataclass

from .types import SettingsType


__all__ = 'merge_dicts', 'resolve_dict', 'merge_settings_reducer',


def merge_dicts(a: dict, b: dict) -> dict:
    return {**b, **a}


def resolve_dict(settings: SettingsType) -> dict:
    """Transforms some internal types like nametuples and dataclasses
    into a plain dict.
    """
    if hasattr(settings, '_asdict'):
        return settings._asdict()

    if is_dataclass(settings):
        return dict(settings.__dict__)

    return settings


def merge_settings_reducer(a: SettingsType, b: SettingsType) -> dict:
    return merge_dicts(resolve_dict(a), resolve_dict(b))
